from onto.attrs.attribute import ReferenceAttribute
reference = ReferenceAttribute
from .store import Store
from .gallery import Gallery
from .struct import struct_ref
from .snapshot_container import SnapshotContainer
from .business_property_store import to_ref
