from .domain_model import DomainModelSource as domain_model, \
    DomainModelTransactionalSource as domain_model_transactional
from .firestore import FirestoreSource as firestore
from .view_model import ViewModelSource as view_model
from .rest import RestViewModelSource as rest

